# This is a placeholder for the tool which is copied from snapd-testing-tools project
# The tool is copied from the tests/lib/external/snapd-testing-tools/tools directory
# while the project is being prepared
echo "placeholder called, source tree not properly setup"
exit 1
