#!/bin/sh

echo "App launched"

cat << EOF > "$SNAP_USER_DATA"/launch-data.txt
args=$*
DISPLAY=${DISPLAY}
WAYLAND_DISPLAY=${WAYLAND_DISPLAY}
XDG_CURRENT_DESKTOP=${XDG_CURRENT_DESKTOP}
XDG_SESSION_DESKTOP=${XDG_SESSION_DESKTOP}
XDG_SESSION_TYPE=${XDG_SESSION_TYPE}
EOF
